/* 
   Graphics library for GGI. Events for AAlib.

   Copyright (C) 1997 ANdreas Beck - becka@sunserver1.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "aavisual.h"
#include "ggi-dl.h"

/************************** GGI Helpers ***************************/
#define MAX_QUEUE 256

struct _queue {
	int cnt;
	ggi_event buff[MAX_QUEUE];
	int head;
	int tail;
};

typedef struct ev_queue {
	ggi_event_mask seen;
	int seencnt[sizeof(ggi_event_mask)*8];
	struct _queue key;	/* emKeyboard events */
	struct _queue ptr;	/* emPointer events */
	struct _queue val;	/* emValuator events */
} ev_queue;

/* Set up an inital, blank queue if none is there
 */
static ev_queue *_ggiEvQueueSetup(ggi_visual *vis)
{
	ev_queue *evq;

	if (VIS_EVQUEUE(vis)==NULL) {
		evq = (ev_queue *)malloc(sizeof(ev_queue));
		memset(evq,0,sizeof(ev_queue));

		VIS_EVQUEUE(vis) = evq;
	} else {
		evq=(ev_queue *)(VIS_EVQUEUE(vis));
	}

	return evq;
}

static ggi_event_mask _ggiEvQueueSeen(ggi_visual *vis,ggi_event_mask mask)
{
	ev_queue *evq;
	ggi_event_mask seen;

	evq=_ggiEvQueueSetup(vis);
	seen=(evq->seen & mask);

	return seen;
}

/* Add an event. Return 0 on success, or drop of unrecogniced,
 * and return -1 is the buffer is full
 */
static int _ggiEvQueueAdd(ggi_visual *vis,ggi_event *ev)
{
	ev_queue *evq=_ggiEvQueueSetup(vis);
	struct _queue *qp;
	ggi_event_mask evm = 1 << ev->any.type;
	
	if (evm & emKeyboard) 
		qp = &(evq->key);
	else if (evm & emPointer)
		qp = &(evq->ptr);
	else if (evm & emValuator)
		qp = &(evq->ptr);
	else {
		return 0;
	}

	if (qp->cnt > MAX_QUEUE)  {
		return -1;
	}

	/* Add the event, and mark that we've seen it.. */
	memcpy(&(qp->buff[qp->head]),ev,ev->any.size);
	qp->head = (qp->head + 1) % MAX_QUEUE;
	qp->cnt++;

	evq->seencnt[ev->any.type]++;
	evq->seen |= evm;

	return 0;
}

static int _ggiEvQueueRelease(ggi_visual *vis,ggi_event *ev,ggi_event_mask mask)
{
	ev_queue *evq=_ggiEvQueueSetup(vis);
	ggi_event_mask evm;
	struct _queue *qp;
	int size;
	uint32 t_key,t_ptr,t_val;
	ggi_event *e_min=NULL,*e_key=NULL,*e_ptr=NULL,*e_val=NULL;
	
	evm=mask & evq->seen;

	/* Got nothing.. */
	if (evm==0) {
		return 0;
	}

	/* Max timestamp.. */
	t_key=t_ptr=t_val=0xFFFFFFFF;

	if (evq->key.cnt != 0) {
		e_key = &(evq->key.buff[evq->key.tail]);
		t_key = e_key->any.time;
	}
	if (evq->ptr.cnt != 0) {
		e_ptr = &(evq->ptr.buff[evq->ptr.tail]);
		t_ptr = e_ptr->any.time;
	}
	if (evq->val.cnt != 0) {
		e_val = &(evq->val.buff[evq->val.tail]);
		t_val = e_val->any.time;
	}
	
	if (t_key < t_ptr) {
		if (t_val < t_key) {
			e_min = e_val;
			qp = &(evq->val);
		} else {
			e_min = e_key;
			qp = &(evq->key);
		}
	} else if (t_val < t_ptr) {
		e_min = e_val;
		qp = &(evq->val);
	} else {
		e_min = e_ptr;
		qp = &(evq->ptr);
	}

	/* Process desired event */
	qp->tail = (qp->tail + 1) % MAX_QUEUE;
	qp->cnt --;
	evq->seencnt[e_min->any.type]--;
	if (evq->seencnt[e_min->any.type]==0)
		evq->seen &= ~(1 << e_min->any.type);

	memcpy(ev,e_min,e_min->any.size);
	size=ev->any.size;

	return size;
}

/************************** GGI Functions ***************************/
/* Event Handling */
ggi_event_mask GGIeventpoll(ggi_visual_t vis,ggi_event_mask mask,
			    struct timeval *t)
{

#define BUFFER_SIZE (sizeof(ggi_event)*4)
	uint n;
	int x,y,b;
	static int lx=0;
	static int ly=0;
	static int lb=0;
	static uint lastkey=0,lastkeyticks=0;
	ggi_event_mask evmask;
	ggi_event ggievent;
	struct AAhooks *aahook=VIS_PRIVATE(vis);

DPRINT("GGIeventpoll(%p,0x%.8x",vis,mask);
if (t==NULL) {
	DPRINT(",NULL)\n");
} else {
	DPRINT(",{%d,%d})\n",t->tv_sec,t->tv_usec);
}

	if (! aahook->context) return 0;
	evmask=_ggiEvQueueSeen(vis,mask);
	if (evmask!=0)  
		return evmask;

	do {
		ggievent.any.focus=0;
		ggievent.any.device=0;
		ggievent.any.time=0;	/* FIXME ! */
		ggievent.any.type=0;

		n=aa_getevent(aahook->context, 0);
		DPRINT("got event %x\n",n);

		if (n==AA_MOUSE)
		{
		  aa_getmouse(aahook->context, &x, &y, &b );
		  ggievent.any.type=evPtrRelative;
		  ggievent.pmove.x=x-lx;lx=x;
		  ggievent.pmove.y=y-lx;ly=y;
		  if (b!=lb)
		  {
		    if ( (b^lb) & b ) {
			  ggievent.any.type=evPtrButtonPress;
			  ggievent.pbutton.state=b;
			  ggievent.pbutton.button=b^lb;
			  lb=b;
		    } else {
			  ggievent.any.type=evPtrButtonRelease;
			  ggievent.pbutton.state=b;
			  ggievent.pbutton.button=b^lb;
			  lb=b;
		    }
		  }
		}
		else if ( n==AA_NONE ) 
		{
			/* ok, if there was a key down, it's now up */
			if (lastkey)
			{
				if(lastkeyticks > 3) /* lastkeyticks - kludge */
				{
	/* fprintf(stderr, "Implied release event %x %c %d, %x %c %d\n", n,n,n,
			lastkey, lastkey, lastkey); */

					ggievent.any.type=evKeyRelease;
					ggievent.key.sym =U(lastkey);
					ggievent.key.code=lastkey;
					lastkey = 0;
					lastkeyticks = 0;
				} else
				{
					lastkeyticks++;
				}
			}
			else
			{
				/* do nothing */
			}
		} 
		else if ( n==AA_RESIZE || n==AA_UNKNOWN )
		{
		  /* Do nothing - not implemented. */
		}
		else if (n>=1&&n<=AA_RELEASE)
		{
			/* ok, check lastkey */
			if(lastkey == 0)
			{
				/* fprintf(stderr, "Press event %x %c %d\n", n,n,n);*/
				ggievent.any.type=evKeyPress;
				ggievent.key.sym=U(n);
				ggievent.key.code=n;
				lastkey = n;
				lastkeyticks = 0;
			}
			else if(lastkey == n)
			{
				/* do nothing - key still down*/
				lastkeyticks++;
			}
			else
			{
				/* whoops!  different key!  We send a release,
				   then we catch the keypress next time around.
				   ** KLUDGE ** */
				/* fprintf(stderr, "Forced release event %x %c %d, %x %c %d\n", n,n,n,
						lastkey,lastkey,lastkey);*/

				ggievent.any.type=evKeyRelease;
				ggievent.key.sym =U(lastkey);
				ggievent.key.code=lastkey;
				lastkey = 0;
				lastkeyticks = 0;
			}
		} 
		else if (n>AA_RELEASE)
		{
			/* AAlib apparently doesn't send release events, just
			 * press events.  As long as the key is down, it returns
			 * a keypress.  When you let it up, it returns nothing.
			 * so, we store the event in lastkey and return *nothing*
			 * if the key is still down. See above */
			/* fprintf(stderr, "Release event: %x %c %d\n", n,n,n); */

			n &= ~AA_RELEASE;
			ggievent.any.type=evKeyRelease;
			ggievent.key.sym =U(n);
			ggievent.key.code=n;
		} 

		if (ggievent.any.type==evKeyPress || 
		    ggievent.any.type==evKeyRelease)
			switch(ggievent.key.code)
			{ 
			  case AA_UP:		ggievent.key.sym=U(K_UP);break;
			  case AA_DOWN:		ggievent.key.sym=U(K_DOWN);break;
			  case AA_LEFT:		ggievent.key.sym=U(K_LEFT);break;
			  case AA_RIGHT:	ggievent.key.sym=U(K_RIGHT);break;
			  case AA_BACKSPACE:	ggievent.key.sym=U(8);break;
			  case AA_ESC:	   	ggievent.key.sym=U(27);break;
			}
		
		switch(ggievent.any.type) {
			case evKeyPress:
			case evKeyRelease:
				ggievent.any.size=sizeof(ggi_key_event);
				ggievent.key.effect=0;
				ggievent.key.normal=0;
				ggievent.key.locked=0;
				ggievent.key.sticky=0;
				break;
			case evPtrRelative:
				ggievent.any.size=sizeof(ggi_pmove_event);
				ggievent.pmove.effect=0;
				ggievent.pmove.normal=0;
				ggievent.pmove.locked=0;
				ggievent.pmove.sticky=0;
				break;
			case evPtrButtonPress:
			case evPtrButtonRelease:
				ggievent.any.size=sizeof(ggi_pbutton_event);
				ggievent.pbutton.effect=0;
				ggievent.pbutton.normal=0;
				ggievent.pbutton.locked=0;
				ggievent.pbutton.sticky=0;
				break;
		}

		if (ggievent.any.type) _ggiEvQueueAdd(vis,&ggievent);
		evmask = _ggiEvQueueSeen(vis,mask);

	} while (evmask==0 && t==NULL);

	return evmask;
}

int GGIeventread(ggi_visual_t vis,ggi_event *ev,ggi_event_mask mask)
{
	/* Block if we don't have anything queued... */
	GGIeventpoll(vis,emAll,NULL);
	return _ggiEvQueueRelease(vis,ev,mask);
}
