/*
 *  Input library for GGI.
 *
 *  Copyright (C) 1997 GGI Project
 *
 *  If you do any modifications, I would like you to send diffs to me
 *  to allow for collecting a more and more complete set of drivers and
 *  to improve existing ones.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----------------------------------------------------------------------
 */

#include "gii.h"

gii_descriptor *gii_alloc_descriptor(char *name,enum gii_desctype dt)
{
	gii_descriptor *desc;

	if (NULL==(desc=malloc(sizeof(gii_descriptor)))) return NULL;
	strncpy(desc->name,name,sizeof(desc->name));
	desc->feat=NULL;
	desc->type=dt;

	if (desc->type<DT_LASTPREALLOCED)	/* Now allocate features */
	{
	/* Fixme. Add a list and constructors. */
	}

return desc;
}

void gii_free_descriptor(gii_descriptor *desc)
{
	/* delete all features */
	while(gii_delete_feature(desc,0)==0);
	free(desc);
}

/* Reading and writing descriptors. */

/*
Descriptor "name" type
  Feature  "name" type
    Control type
      Privatedata "..."
        Control type
          Privatedata "..."
            Feature ...
            Descriptor ...
            ...

Example :

Descriptor "thrust-control" DT_AXIS	# controlling the ship speed
  Feature "forward" POS_DIR		# we can go forward or back.
    Control KEY_PRESS			# one method is pressing a key
      Privatedata "dev=KBD1 sym=CRSR_UP 0 100"	# (the key is CRSR-UP).
    Control AXIS_RANGE			# the other moving the joystick
      Privatedata "Joy1 AxY2 330 25"	# axis Y2, value=330 for 0, 25 for 100%
  Feature "backward" POS_DIR		# go back.
    Control KEY_PRESS			# one method is pressing a key
      Privatedata "dev=KBD1 sym=CRSR_DOWN 0 100"	# (the key is CRSR-DOWN).
    Control AXIS_RANGE			# the other moving the joystick
      Privatedata "Joy1 AxY2 330 660"	# axis Y2, value=330 for 0, 660 for 100%

*/
              
/* Write out a descriptor to a human-readable file. */

int gii_write_descriptor(FILE *file,gii_descriptor *desc)
{
	gii_feature *feat;
	gii_control *control;
	char buffer[1024];
	
	fprintf(file,"Descriptor \"%s\" %d\n",desc->name,desc->type);
	
	feat=desc->feat;
	while(feat)
	{
		fprintf(file," Feature \"%s\" %d\n",feat->name,feat->type);
		control=feat->ctrl;
		while(control)
		{
			fprintf(file,"  Control %d\n",control->type);
			control->cops->write_pvtdata(control,buffer,sizeof(buffer));
			fprintf(file,"   Privatedata %s\n",buffer);
			control=control->next;
		}
		feat=feat->next;
	}

/* returns?? */
}

static char *getline(char *buffer,int size,FILE *file)
{
	char *rc;
	do {
		rc=fgets(buffer,size,file);
	} while(rc && (*buffer=='\n'||*buffer=='#'));
}

/* Read such a descriptor back. It will be allocated automatically. If the returned
 * Pointer is NULL, then no valid GII descriptor was found at the current file
 * position. In this case, the RC is nonzero, too.
 */
int gii_read_descriptor (FILE *file,gii_descriptor **desc)
{
	gii_feature feat;
	gii_control control;
	char buffer[1024];
	
	(*desc)=gii_alloc_descriptor("",DT_PRIVATE);	/* This guarantees no alloced features. */
	getline(buffer,sizeof(buffer),file);
	sscanf(buffer,"Descriptor \"%[^\"]\" %d\n",(*desc)->name,(*desc)->type);
	
	while(1)
	{
		feat.control=NULL;
		fscanf(file," Feature \"%[^\"]\" %d\n",feat.name,&feat.type);
		while(1)
		{
			fscanf(file,"  Control %d\n",control->type);
			fscanf(file,"   Privatedata %s\n",buffer);
		}
	}
}

int    gii_autoconfig_control(gii_descriptor *desc,int featurenumber,int controlnumber);
uint32 gii_check             (ggi_event *event,gii_descriptor *desc);


gii_feature *gii_find_feature(gii_descriptor *desc,int featurenumber)
{
	gii_feature *feat;

	feat=desc->feat;
	if (!feat) return NULL;
	while(featurenumber--)
	{ 
		feat=feat->next;
		if (!feat) return NULL;
	}
	return feat;
}

gii_control *gii_find_control(gii_feature *feat,int controlnumber)
{
	gii_control *cont;

	if (!feat) return NULL;
	cont=feat->ctrl;
	if (!cont) return NULL;
	while(controlnumber--)
	{ 
		cont=cont->next;
		if (!cont) return NULL;
	}
	return cont;
}

int gii_delete_control    (gii_descriptor *desc,int featurenumber,int controlnumber)
{
	gii_feature *feat;
	gii_control *ctrl,**ctrlptr;

	feat=gii_find_feature(desc,featurenumber);
	if (!feat) return -1;
	ctrl=*(ctrlptr=&(feat->ctrl));
	if (!ctrl) return -1;
	while(controlnumber--)
	{ 
		ctrl=*(ctrlptr=(&ctrl->next));
		if (!ctrl) return -1;
	}
	*ctrlptr=ctrl->next;
	free(feat);
	return 0;
}

int gii_delete_feature(gii_descriptor *desc,int featurenumber)
{
	gii_feature *feat,**featptr;

	while(gii_delete_control(desc,featurenumber,0)==0);

	feat=*(featptr=(&desc->feat));
	if (!feat) return -1;
	while(featurenumber--)
	{ 
		feat=*(featptr=(&feat->next));
		if (!feat) return -1;
	}
	*featptr=feat->next;
	free(feat);
	return 0;
}

int gii_get_feature_name(gii_descriptor *desc,int featurenumber,
                        char *buffer,int maxlen)
{
	gii_feature *feat;
	feat=gii_find_feature(desc,featurenumber);
	if (!feat) { *buffer='\0'; return -1; }
	strncpy(buffer,feat->name,maxlen);
	return 0;
}

int gii_get_control_name(gii_descriptor *desc,int featurenumber,
                        int control,char *buffer,int maxlen)
{
	gii_control *ctrl;
	ctrl=gii_find_control(gii_find_feature(desc,featurenumber),control);
	if (!ctrl) { *buffer='\0'; return -1; }
	strncpy(buffer,ctrl->name,maxlen);
	return 0;
}

/* Testing
 */

main()
{
	int x,y;
	char buffer[256];
	gii_descriptor *menu;
	
	menu=gii_alloc_descriptor("Main Menu",DT_MENU);
	printf("Allocated gii_descriptor %p for menu.\n",menu);

	for(x=0;;x++)
	{
	  if (gii_get_feature_name(menu,x,buffer,sizeof(buffer))) break;
	  printf("Feature #%d: %s\n",x,buffer);
	  for(y=0;;y++)
	  {
	    if (gii_get_control_name(menu,x,y,buffer,sizeof(buffer))) break;
	    printf(" Control #%d: %s\n",y,buffer);
	  }
	}
	gii_write_descriptor(stdout,menu);
	gii_free_descriptor(menu);
	printf("Freed it.\n");

}
